# pylint: disable=abstract-method

from __future__ import absolute_import

from interface_meta import override

from omniduct.utils.debug import logger

from .base import DatabaseClient


class Neo4jClient(DatabaseClient):
    """
    This Duct connects to a Neo4j graph database server using the `neo4j` python
    library.
    """

    PROTOCOLS = ["neo4j"]
    DEFAULT_PORT = 7687
    DEFAULT_CURSOR_FORMATTER = "raw"

    @override
    @classmethod
    def statement_cleanup(cls, statement):
        return statement  # base statement cleanup assumes SQL

    @override
    def _init(self):
        self.__driver = None

    # Connection
    @override
    def _connect(self):
        from neo4j.v1 import GraphDatabase

        logger.info("Connecting to Neo4J graph database ...")
        auth = (self.username, self.password) if self.username else None
        # pylint: disable-next=attribute-defined-outside-init
        self.__driver = GraphDatabase.driver(
            f"bolt://{self.host}:{self.port}", auth=auth
        )  # TODO: Add kerberos support

    @override
    def _is_connected(self):
        return hasattr(self, "__driver") and self.__driver is not None

    @override
    def _disconnect(self):
        logger.info("Disconnecting from Neo4J graph database ...")
        try:
            self.__driver.close()
        except:  # pylint: disable=bare-except
            pass
        self.__driver = None  # pylint: disable=attribute-defined-outside-init

    # Querying
    @override
    def _execute(self, statement, cursor, wait, session_properties):
        with self.__driver.session() as session:
            result = session.run(statement)

        # hacky: make the result look like a cursor
        result.close = result.detach
        result.fetchall = result.records
        return result

    @override
    def _table_exists(self, table, **kwargs):
        raise RuntimeError("tables do not apply to the Neo4J graph database")

    @override
    def _table_drop(self, table, **kwargs):
        raise RuntimeError("tables do not apply to the Neo4J graph database")

    @override
    def _table_desc(self, table, **kwargs):
        raise RuntimeError("tables do not apply to the Neo4J graph database")

    @override
    def _table_head(self, table, n=10, **kwargs):
        raise RuntimeError("tables do not apply to the Neo4J graph database")

    @override
    def _table_list(self, namespace, **kwargs):
        raise RuntimeError("tables do not apply to the Neo4J graph database")

    @override
    def _table_props(self, table, **kwargs):
        raise RuntimeError("tables do not apply to the Neo4J graph database")



import datetime
import functools
from abc import abstractmethod

import dateutil
import pandas
import yaml
from decorator import decorator
from interface_meta import inherit_docs

from omniduct.duct import Duct
from omniduct.utils.config import config
from omniduct.utils.debug import logger
from omniduct.utils.decorators import function_args_as_kwargs, require_connection

from ._serializers import PickleSerializer

config.register(
    "cache_fail_hard",
    description="Raise an exception if a cache fails to save (otherwise errors are logged and suppressed).",
    default=False,
)


def cached_method(
    key,
    namespace=lambda self, kwargs: (
        self.cache_namespace or f"{self.__class__.__name__}.{self.name}"
    ),
    cache=lambda self, kwargs: self.cache,
    use_cache=lambda self, kwargs: kwargs.pop("use_cache", True),
    renew=lambda self, kwargs: kwargs.pop("renew", False),
    serializer=lambda self, kwargs: PickleSerializer(),
    metadata=lambda self, kwargs: None,
):
    """
    Wrap a method of a `Duct` class and add caching capabilities.

    All arguments of this function are expected to be functions taking two
    arguments: a reference to current instance of the class (`self`) and a
    dictionary of arguments passed to the function (`kwargs`).

    Args:
        key (function -> str): The key under which the value returned by the
            wrapped function should be stored.
        namespace (function -> str): The namespace under which the key should be
            stored (default: `"<duct class name>.<duct instance name>"`).
        cache (function -> Cache): The instance of cache via which to store the
            output of the wrapped function (default: `self.cache`).
        use_cache (function -> bool): Whether or not to use the caching
            functionality (default: `True`).
        renew (function -> bool): Whether to renew the stored cache, overriding
            if a value has already been stored (default: `False`).
        serializer (function -> Serializer): The `Serializer` subclass to use
            when storing the return object (default: `PickleSerializer`).
        metadata (function -> None, dict): A dictionary of additional metadata
            to be stored alongside the wrapped function's output
            (default: `None`).

    Returns:
        object: The (potentially cached) object returned when calling the
            wrapped function.

    Raises:
        Exception: If cache fails to store the output of the wrapped function,
            and the omniduct configuration key `cache_fail_hard` is `True`, then
            the underlying exceptions raised by the Cache instance will be
            reraised.
    """

    @decorator
    def wrapped(method, self, *args, **kwargs):
        kwargs = function_args_as_kwargs(method, self, *args, **kwargs)
        kwargs.pop("self")

        _key = key(self, kwargs)
        _namespace = namespace(self, kwargs)
        _cache = cache(self, kwargs)
        _use_cache = use_cache(self, kwargs)
        _renew = renew(self, kwargs)
        _serializer = serializer(self, kwargs)
        _metadata = metadata(self, kwargs)

        if _cache is None or not _use_cache:
            return method(self, **kwargs)

        if (
            _cache.has_key(_key, namespace=_namespace) and not _renew
        ):  # noqa: has_key is not of a dictionary here
            try:
                return _cache.get(_key, namespace=_namespace, serializer=_serializer)
            except Exception as e:  # pylint: disable=broad-exception-caught
                logger.warning(
                    "Failed to retrieve results from cache [%s]. Renewing the cache...",
                    e,
                )
                if config.cache_fail_hard:
                    raise
            finally:
                logger.caveat("Loaded from cache")

        # Renewing/creating cache
        value = method(self, **kwargs)
        if value is None:
            logger.warning("Method value returned None. Not saving to cache.")
            return None

        try:
            _cache.set(
                _key,
                value=value,
                namespace=_namespace,
                serializer=_serializer,
                metadata=_metadata,
            )
            # Return from cache every time, just in case serialization operation was
            # destructive (e.g. reading from cursors)
            return _cache.get(_key, namespace=_namespace, serializer=_serializer)
        except:  # pylint: disable=bare-except
            logger.warning(
                "Failed to save results to cache. If needed, please save them manually."
            )
            if config.cache_fail_hard:
                raise
            return value  # As a last resort, return value object (which could be mutated by serialization).

    return wrapped


class Cache(Duct):
    """
    An abstract class providing the common API for all cache clients.
    """

    DUCT_TYPE = Duct.Type.CACHE

    @inherit_docs("_init", mro=True)
    def __init__(self, **kwargs):  # pylint: disable=super-init-not-called
        Duct.__init_with_kwargs__(self, kwargs)
        self._init(**kwargs)

    @abstractmethod
    def _init(self):
        pass

    # Data insertion and retrieval

    @require_connection
    def set(self, key, value, namespace=None, serializer=None, metadata=None):
        """
        Set the value of a key.

        Args:
            key (str): The key for which `value` should be stored.
            value (object): The value to be stored.
            namespace (str, None): The namespace to be used.
            serializer (Serializer): The `Serializer` subclass to use for the
                serialisation of value into the cache. (default=PickleSerializer)
            metadata (dict, None): Additional metadata to be stored with the value
                in the cache. Values must be serializable via `yaml.safe_dump`.
        """
        namespace, key = self._namespace(namespace), self._key(key)
        serializer = serializer or PickleSerializer()
        try:
            with self._get_stream_for_key(
                namespace,
                key,
                f"data{serializer.file_extension}",
                mode="wb",
                create=True,
            ) as fh:
                serializer.serialize(value, fh)
            self.set_metadata(key, metadata, namespace=namespace, replace=True)
        except:  # pylint: disable=bare-except
            self.unset(key, namespace=namespace)
            raise

    @require_connection
    def set_metadata(self, key, metadata, namespace=None, replace=False):
        """
        Set the metadata associated with a stored key, creating the key if it
        is missing.

        Args:
            key (str): The key for which `value` should be stored.
            metadata (dict, None): Additional/override metadata to be stored
                for `key` in the cache. Values must be serializable via
                `yaml.safe_dump`.
            namespace (str, None): The namespace to be used.
            replace (bool): Whether the provided metadata should entirely
                replace any existing metadata, or just update it. (default=False)
        """
        namespace, key = self._namespace(namespace), self._key(key)
        if replace:
            orig_metadata = {"created": datetime.datetime.utcnow()}
        else:
            orig_metadata = self.get_metadata(key, namespace=namespace)

        orig_metadata.update(metadata or {})

        with self._get_stream_for_key(
            namespace, key, "metadata", mode="w", create=True
        ) as fh:
            yaml.safe_dump(orig_metadata, fh, default_flow_style=False)

    @require_connection
    def get(self, key, namespace=None, serializer=None):
        """
        Retrieve the value associated with the nominated key from the cache.

        Args:
            key (str): The key for which `value` should be retrieved.
            namespace (str, None): The namespace to be used.
            serializer (Serializer): The `Serializer` subclass to use for the
                deserialisation of value from the cache. (default=PickleSerializer)

        Returns:
            object: The (appropriately deserialized) object stored in the cache.
        """
        namespace, key = self._namespace(namespace), self._key(key)
        serializer = serializer or PickleSerializer()
        if not self._has_key(namespace, key):
            raise KeyError(f"{key} (namespace: {namespace})")
        try:
            with self._get_stream_for_key(
                namespace,
                key,
                f"data{serializer.file_extension}",
                mode="rb",
                create=False,
            ) as fh:
                return serializer.deserialize(fh)
        finally:
            self.set_metadata(
                key,
                namespace=namespace,
                metadata={"last_accessed": datetime.datetime.utcnow()},
            )

    @require_connection
    def get_bytecount(self, key, namespace=None):
        """
        Retrieve the number of bytes used by a stored key.

        This bytecount may or may not include metadata storage, depending on
        the backend.

        Args:
            key (str): The key for which to extract the bytecount.
            namespace (str, None): The namespace to be used.

        Returns:
            int: The number of bytes used by the stored value associated with
                the nominated key and namespace.
        """
        namespace, key = self._namespace(namespace), self._key(key)
        if not self._has_key(namespace, key):
            raise KeyError(f"{key} (namespace: {namespace})")
        return self._get_bytecount_for_key(namespace, key)

    @require_connection
    def get_metadata(self, key, namespace=None):
        """
        Retrieve metadata associated with the nominated key from the cache.

        Args:
            key (str): The key for which to extract metadata.
            namespace (str, None): The namespace to be used.

        Returns:
            dict: The metadata associated with this namespace and key.
        """
        namespace, key = self._namespace(namespace), self._key(key)
        if not self._has_key(namespace, key):
            raise KeyError(f"{key} (namespace: {namespace})")
        try:
            with self._get_stream_for_key(
                namespace, key, "metadata", mode="r", create=False
            ) as fh:
                return yaml.safe_load(fh)
        except:  # pylint: disable=bare-except
            return {}

    @require_connection
    def unset(self, key, namespace=None):
        """
        Remove the nominated key from the cache.

        Args:
            key (str): The key which should be unset.
            namespace (str, None): The namespace to be used.
        """
        namespace, key = self._namespace(namespace), self._key(key)
        if not self._has_key(namespace, key):
            raise KeyError(f"{key} (namespace: {namespace})")
        self._remove_key(namespace, key)

    @require_connection
    def unset_namespace(self, namespace=None):
        """
        Remove an entire namespace from the cache.

        Args:
            namespace (str, None): The namespace to be removed.
        """
        namespace = self._namespace(namespace)
        if not self._has_namespace(namespace):
            raise KeyError(f"namespace: {namespace}")
        self._remove_namespace(namespace)

    # Top-level descriptions

    @property
    @require_connection
    def namespaces(self):
        "list <str,None>: A list of the namespaces stored in the cache."
        return self._get_namespaces()

    @require_connection
    def has_namespace(self, namespace=None):
        """
        Check whether the cache has the nominated namespace.

        Args:
            namespace (str,None): The namespace for which to check for existence.

        Returns:
            bool: Whether the cache has the nominated namespaces.
        """
        namespace = self._namespace(namespace)
        return self._has_namespace(namespace)

    @require_connection
    def keys(self, namespace=None):
        """
        Collect a list of all the keys present in the nominated namespaces.

        Args:
            namespace (str,None): The namespace from which to extract all of the
                keys.

        Returns:
            list<str>: The keys stored in the cache for the nominated namespace.
        """
        namespace = self._namespace(namespace)
        return self._get_keys(namespace)

    @require_connection
    def has_key(self, key, namespace=None):
        """
        Check whether the cache as a nominated key.

        Args:
            key (str): The key for which to check existence.
            namespace (str,None): The namespace from which to extract all of the
                keys.

        Returns:
            bool: Whether the cache has a value for the nominated namespace and
                key.
        """
        namespace, key = self._namespace(namespace), self._key(key)
        return self._has_key(namespace, key)

    def get_total_bytecount(self, namespaces=None):
        """
        Retrieve the total number of bytes used by the cache.

        This method iterates over all (nominated) namespaces and the keys
        therein, summing the result of `.get_bytecount(...)` on each.

        Args:
            namespaces (list<str,None>): The namespaces to which the bytecount
                should be restricted.

        Returns:
            int: The total number of bytes used by the nominated namespaces.
        """
        total_bytes = 0

        if namespaces is None:
            namespaces = self.namespaces

        for namespace in namespaces:
            for key in self.keys(namespace=namespace):
                total_bytes += self.get_bytecount(key, namespace=namespace)

        return total_bytes

    def describe(self, namespaces=None):
        """
        Return a pandas DataFrame showing all keys and their metadata.

        Args:
            namespaces (list<str,None>): The namespaces to which the summary
                should be restricted.

        Returns:
            pandas.DataFrame: A representation of keys in the cache. Will include
                at least the following columns: ['bytes', 'namespace', 'key',
                'created', 'last_accessed']. Any additional metadata for keys
                will be appended to these columns.
        """
        out = []

        if namespaces is None:
            namespaces = self.namespaces

        for namespace in namespaces:
            for key in self.keys(namespace=namespace):
                usage = {
                    "bytes": self.get_bytecount(key, namespace=namespace),
                    "namespace": namespace,
                    "key": key,
                    "created": None,
                    "last_accessed": None,
                }
                usage.update(self.get_metadata(key, namespace=namespace))
                out.append(usage)

        required_columns = ["bytes", "namespace", "key", "created", "last_accessed"]
        if out:
            df = pandas.DataFrame(out)
            order = required_columns + sorted(
                set(df.columns).difference(required_columns)
            )
            return df.sort_values("last_accessed", ascending=False).reset_index(
                drop=True
            )[order]

        return pandas.DataFrame(data=[], columns=required_columns)

    # Cache pruning

    def prune(
        self,
        namespaces=None,
        max_age=None,
        max_bytes=None,
        total_bytes=None,
        total_count=None,
    ):
        """
        Remove keys from the cache in order to satisfy nominated constraints.

        Args:
            namespaces (list<str, None>): The namespaces to consider for pruning.
            max_age (None, int, timedelta, relativedelta, date, datetime): The
                number of days, a timedelta, or a relativedelta, indicating the
                maximum age of items in the cache (based on last accessed date).
                Deltas are expected to be positive.
            max_bytes (None, int): The maximum number of bytes for *each* key,
                allowing the pruning of larger keys.
            total_bytes (None, int): The total number of bytes for the entire
                cache. Keys will be removed from least recently accessed to most
                recently accessed until the constraint is satisfied. This
                constraint will be applied after max_age and max_bytes.
            total_count (None, int): The maximum number of items to keep in the
                cache. Keys will be removed from least recently accessed to most
                recently accessed until the constraint is satisfied. This
                constraint will be applied after max_age and max_bytes.
        """
        usage = self.describe(namespaces=namespaces)
        if (
            usage.shape[0] == 0
        ):  # Abort early if the cache is empty (and hence has no index, which would cause problems later on)
            return

        constraints = []

        # Unset keys according to per-key constraints
        if max_age is not None:
            if isinstance(max_age, int):
                max_age = datetime.timedelta(max_age)
            if isinstance(
                max_age, (datetime.timedelta, dateutil.relativedelta.relativedelta)
            ):
                max_age = datetime.datetime.now() - max_age
            if not isinstance(max_age, (datetime.datetime, datetime.date)):
                raise ValueError(
                    f"Invalid type specified for `max_age`: {repr(max_age)}"
                )
            constraints.append(usage.last_accessed < max_age)

        if max_bytes is not None:
            if not isinstance(max_bytes, int):
                raise ValueError(
                    f"Invalid type specified for `max_bytes`: {repr(max_bytes)}"
                )
            constraints.append(usage.bytes > max_bytes)

        if constraints:
            to_unset = usage[functools.reduce(lambda x, y: x | y, constraints, False)]
            for _, row in to_unset.iterrows():
                logger.info(
                    f"Unsetting key '{row.key}' (namespace: '{row.namespace}')..."
                )
                self.unset(row.key, namespace=row.namespace)

        # Unset keys according to global constraints
        if total_bytes is not None or total_count is not None:
            if total_bytes is not None and not isinstance(total_bytes, int):
                raise ValueError(
                    f"Invalid type specified for `total_bytes`: {repr(total_bytes)}"
                )
            if total_count is not None and not isinstance(total_count, int):
                raise ValueError(
                    f"Invalid type specified for `total_count`: {repr(total_bytes)}"
                )
            usage = self.describe(namespaces=namespaces).assign(
                cum_bytes=lambda x: x.bytes.cumsum()
            )

            unset_index = total_count if total_count is not None else len(usage)
            if total_bytes is not None:
                unset_index = min(
                    unset_index, usage.cum_bytes.searchsorted(total_bytes, side="right")
                )
            for _, row in usage.loc[unset_index:].iterrows():
                logger.info(
                    f"Unsetting key '{row.key}' (namespace: '{row.namespace}')..."
                )
                self.unset(row.key, namespace=row.namespace)

    # Methods for subclasses to implement

    def _namespace(self, namespace):
        return namespace

    def _key(self, key):
        return key

    @abstractmethod
    def _get_namespaces(self):
        raise NotImplementedError

    def _has_namespace(self, namespace):
        return namespace in self._get_namespaces()

    @abstractmethod
    def _remove_namespace(self, namespace):
        raise NotImplementedError

    @abstractmethod
    def _get_keys(self, namespace):
        raise NotImplementedError

    def _has_key(self, namespace, key):
        return key in self._get_keys(namespace=namespace)

    @abstractmethod
    def _remove_key(self, namespace, key):
        raise NotImplementedError

    @abstractmethod
    def _get_bytecount_for_key(self, namespace, key):
        raise NotImplementedError

    @abstractmethod
    def _get_stream_for_key(self, namespace, key, stream_name, mode, create):
        pass



import importlib
import pkgutil
import sys


def import_submodules(package_name):
    """Import all submodules of a module, recursively
    :param package_name: Package name
    :type package_name: str
    :rtype: dict[types.ModuleType]
    """
    package = sys.modules[package_name]
    return {
        name: importlib.import_module(package_name + "." + name)
        for loader, name, is_pkg in pkgutil.walk_packages(package.__path__)
    }



from pandas.io.sql import SQLTable, SQLDatabase


def to_sql(df, name, schema, con, index, if_exists, mode="default", **kwargs):
    """
    Override the default `pandas.to_sql` method to allow for insertion of
    multiple rows of data at once. This is derived from the upstream patch at
    https://github.com/pandas-dev/pandas/pull/21401, and can be deprecated
    once it is merged and released in a new version of `pandas`.
    """
    assert mode in ("default", "multi"), f"unexpected `to_sql` mode {mode}"
    if mode == "default":
        return df.to_sql(
            name=name,
            schema=schema,
            con=con,
            index=index,
            if_exists=if_exists,
            **kwargs,
        )

    nrows = len(df)
    if nrows == 0:
        return None

    chunksize = kwargs.get("chunksize", nrows)
    if chunksize == 0:
        raise ValueError("chunksize argument should be non-zero")
    chunks = int(nrows / chunksize) + 1

    pd_sql = SQLDatabase(con)
    pd_table = SQLTable(
        name,
        pd_sql,
        frame=df,
        index=index,
        if_exists=if_exists,
        index_label=kwargs.get("insert_label"),
        schema=schema,
        dtype=kwargs.get("dtype"),
    )
    pd_table.create()
    keys, data_list = pd_table.insert_data()

    with pd_sql.run_transaction() as conn:
        for i in range(chunks):
            start_i = i * chunksize
            end_i = min((i + 1) * chunksize, nrows)
            if start_i >= end_i:
                break

            chunk_iter = zip(*[arr[start_i:end_i] for arr in data_list])
            data = [dict(zip(keys, row)) for row in chunk_iter]
            conn.execute(pd_table.table.insert(data))  # multivalues insert
    return None



import pickle

import pandas


class Serializer:
    @property
    def file_extension(self):
        return ""

    def serialize(self, obj, fh):
        raise NotImplementedError

    def deserialize(self, fh):
        raise NotImplementedError


class BytesSerializer(Serializer):
    @property
    def file_extension(self):
        return ".bytes"

    def serialize(self, obj, fh):
        assert isinstance(
            obj, bytes
        ), "BytesSerializer requires incoming data be already encoded into a bytestring."
        fh.write(obj)

    def deserialize(self, fh):
        return fh.read()


class PickleSerializer(Serializer):
    @property
    def file_extension(self):
        return ".pickle"

    def serialize(self, obj, fh):
        return pickle.dump(obj, fh)

    def deserialize(self, fh):
        return pickle.load(fh)


class PandasSerializer(Serializer):
    @property
    def file_extension(self):
        return ".pandas"

    def serialize(self, obj, fh):
        return pandas.to_pickle(obj, fh, compression=None)

    def deserialize(self, fh):
        return pandas.read_pickle(fh, compression=None)



# flake8: noqa

from omniduct.duct import Duct
from omniduct.registry import DuctRegistry
from omniduct.utils.config import config
from omniduct.utils.debug import logger

from . import protocols
from ._version import __author__, __author_email__, __version__, __logo__, __docs_url__


def about():
    from collections import OrderedDict
    from .utils.about import show_about

    return show_about(
        "Omniduct",
        version=__version__,
        logo=__logo__,
        maintainers=OrderedDict(
            zip(
                [a.strip() for a in __author__.split(",")],
                [a.strip() for a in __author_email__.split(",")],
            )
        ),
        attributes={
            "Documentation": __docs_url__,
        },
        description="""
        Omniduct provides uniform interfaces for connecting to and extracting data
        from a wide variety of (potentially remote) data stores (including HDFS,
        Hive, Presto, MySQL, etc).
        """,
        endorse_omniduct=False,
    )



from collections import OrderedDict

from omniduct.utils.about import show_about as _show_about

from ._version import __author__, __author_email__, __version__  # noqa: F401
from .services import config, logger, registry  # noqa: F401


# Expose services via a proxy object
services = registry.get_proxy(by_kind=True)

# For convenience, add some services to the top-level module
registry.populate_namespace(
    namespace=globals(),
    names=['presto']
)


def about():
    """
    Show information about this package.
    """
    _show_about(
        name='Example Wrapper',
        version=__version__,
        maintainers=OrderedDict(zip(
            [a.strip() for a in __author__.split(',')],
            [a.strip() for a in __author_email__.split(',')]
        )),
        description="""
        A simple example wrapper around Omniduct for pre-configuring services
        in the context of an organisation.
        """
    )



from setuptools import find_packages, setup

# Extract version information from Omniduct _version.py
version_info = {}
with open('example_wrapper/_version.py') as version_file:
    exec(version_file.read(), version_info)

setup(
    # Package metadata
    name="example_wrapper",
    version=version_info['__version__'],
    author=version_info['__author__'],
    author_email=version_info['__author_email__'],
    url='http://git.company.com/example_wrapper',
    description="Exposes Company.com services locally.",
    license='Internal',

    # Package details
    packages=find_packages(),
    include_package_data=True,

    # Dependencies
    install_requires=version_info['__dependencies__']
)



import base64
import textwrap
import urllib.parse

import jinja2

import omniduct


ABOUT_TEMPLATE_HTML = """
{%- if logo %}
<div><img src="{{logo}}" style="height: 200px;"/></div>
{%- endif %}
<b>{{name}}</b> v{{version}}<br/>

{%- if maintainers %}
<b>Maintainers:</b> {% for name, email in maintainers.items() %}{% if loop.index0 > 0 %}, {% endif %}<a href="mailto:{{email}}">{{name}}</a>{% endfor %}<br/>
{%- endif %}
{%- for name, value in attributes.items() %}
<b>{{name}}:</b> {% if value.startswith('http://') or value.startswith('https://') %}<a href='{{value}}'>{{value}}</a>{% else %}{{value}}{%endif%}<br />
{%- endfor %}
<b>Description:</b><br/>
{{description.strip()|replace('\n\n','<br/><br/>\n')}}
{%- if endorsements %}<br />
<br />
<b>Built upon:</b>
{%- for endorsement in endorsements %}
<div>
{%- if endorsement['logo'] -%}
<img src="{{endorsement['logo']}}" style="display: inline-block; height: 2em; width: 2em; object-size: contain; vertical-align: middle; margin-right: 0.5em;"/>
{%- else -%}
<span style='display: inline-block; width: 2em; height: 2em; vertical-align: middle; margin-right: 0.5em;'></span>
{%- endif -%}
{{endorsement['name']}} v{{endorsement['version'].strip()}}</div>
{%- endfor %}
{%- endif %}
"""

ABOUT_TEMPLATE_TEXT = """
{{name}}{% if version %} v{{version}}{% endif %}

{%- if maintainers %}
- Maintainers: {% for name, email in maintainers.items() %}{% if loop.index0 > 0 %}, {% endif %}{{name}} <{{email}}>{% endfor %}
{%- endif %}
{%- for name, value in attributes.items() %}
- {{name}}: {{value}}
{%- endfor %}

{{description}}
{% if endorsements %}
Built upon:
{%- for endorsement in endorsements %}
- {{endorsement['name']}} v{{endorsement['version'].strip()}}
{%- endfor %}
{%- endif %}
""".strip()


def show_about(
    name,
    version=None,
    logo=None,
    maintainers=None,
    attributes=None,
    description=None,
    endorsements=None,
    endorse_omniduct=True,
):
    """
    Output information about a project in HTML for notebooks and text otherwise.

    Args:
        name (str): The name of the project.
        version (str, None): The version of the project.
        logo (str, None): A local or remote uri for the project logo.
        maintainers (dict, None): A dictionary mapping name to email address for
            maintainers. If order is important pass an `OrderedDict`.
        attributes (dict, None): A dictionary mapping of project attributes to
            values. If order is important pass an `OrderedDict`.
        description (str, None): A description of the project.
        endorsements (list<dict>): A list of dependencies to highlight encoded
            as dictionaries of form: {'name': ..., 'version': ..., 'logo': ...}.
            Only `name` is required, and endorsements will be sorted by name.
        endorse_omniduct (bool): Whether to include omniduct in the list of
            endorsements (default: True).
    """

    endorsements = endorsements or []
    if endorse_omniduct:
        endorsements.append(
            {
                "name": "Omniduct",
                "version": omniduct.__version__,
                "logo": omniduct.__logo__,
            }
        )
    for endorsement in endorsements:
        endorsement["logo"] = get_image_url(endorsement.get("logo"))
    endorsements = sorted(endorsements, key=lambda x: x["name"])

    context = {
        "name": name,
        "version": version,
        "logo": get_image_url(logo),
        "maintainers": maintainers or {},
        "attributes": attributes or {},
        "description": textwrap.dedent(description).strip() if description else None,
        "endorsements": endorsements,
    }

    try:
        from IPython import get_ipython
        from IPython.display import display, HTML

        ip = get_ipython()
        if ip is not None and ip.has_trait("kernel"):
            return display(HTML(jinja2.Template(ABOUT_TEMPLATE_HTML).render(**context)))
    except:  # pylint: disable=bare-except
        pass

    # Textual fallback if HTML not running in a notebook
    return print(
        textwrap.dedent(jinja2.Template(ABOUT_TEMPLATE_TEXT).render(**context))
    )


def get_image_url(uri):
    """
    Get a base64 URI if uri is a local path or pass through value otherwise.

    This is used to allow rendering of images in notebooks for local files,
    such as project logos.

    Args:
        uri (str, None): The local path of a logo or remote image.

    Returns:
        str: The uri of the image suitable for rendering in a notebook.
    """
    if not uri:
        return None
    parsed = urllib.parse.urlparse(uri)
    if parsed.scheme in ("", "file"):
        with open(parsed.path, "rb") as image:
            return f"data:image/png;base64,{base64.b64encode(image.read()).decode()}"
    return uri



import http.client
import json
import xml.dom.minidom


import requests

from pywebhdfs import errors
from pywebhdfs.webhdfs import (
    PyWebHdfsClient,
    _is_standby_exception,
    _move_active_host_to_head,
)


class OmniductPyWebHdfsClient(PyWebHdfsClient):
    """
    A wrapper around `pywebhdfs.PyWebHdfsClient` to handle redirects requested
    by the namenodes when taking advantage of Omniduct's automatic
    port-forwarding of remote services.
    """

    def __init__(self, remote=None, namenodes=None, **kwargs):
        self.remote = remote
        self.namenodes = namenodes or []

        PyWebHdfsClient.__init__(self, **kwargs)

        if self.namenodes and "path_to_hosts" not in kwargs:
            self.path_to_hosts = [(".*", self.namenodes)]

        # Override base uri
        self.base_uri_pattern = kwargs.get(
            "base_uri_pattern", "http://{host}/webhdfs/v1/"
        ).format(host="{host}")

    @property
    def host(self):
        host = "localhost" if self.remote else self._host
        return f"{host}:{str(self.port)}"

    @host.setter
    def host(self, host):
        self._host = host

    @property
    def port(self):
        if self.remote:
            return self.remote.port_forward(f"{self._host}:{self._port}")
        return self._port

    @port.setter
    def port(self, port):
        self._port = port

    @property
    def namenodes(self):
        if self.remote:
            return [
                f"localhost:{self.remote.port_forward(nn)}" for nn in self._namenodes
            ]
        return self._namenodes

    @namenodes.setter
    def namenodes(self, namenodes):
        self._namenodes = namenodes

    def _make_uri_local(self, uri):
        if not self.remote:
            return uri
        uri = self.remote.get_local_uri(uri)
        return uri

    def get_home_directory(self):
        response = self._resolve_host(
            requests.get, True, "/", operation="GETHOMEDIRECTORY"
        )
        if response.ok:
            return json.loads(response.content)["Path"]
        return "/"

    def _resolve_host(self, req_func, allow_redirect, path, operation, **kwargs):
        """
        This is where the magic happens, and where omniduct handles redirects
        during federation and HA.
        """
        uri_without_host = self._create_uri(path, operation, **kwargs)
        hosts = self._resolve_federation(path)
        for host in hosts:
            uri = uri_without_host.format(host=host)
            try:
                while True:
                    response = req_func(
                        uri,
                        allow_redirects=False,
                        timeout=self.timeout,
                        **self.request_extra_opts,
                    )

                    if (
                        allow_redirect
                        and response.status_code == http.client.TEMPORARY_REDIRECT
                    ):
                        uri = self._make_uri_local(response.headers["location"])
                    else:
                        break

                if (
                    not allow_redirect
                    and response.status_code == http.client.TEMPORARY_REDIRECT
                ):
                    response.headers["location"] = self._make_uri_local(
                        response.headers["location"]
                    )

                if not _is_standby_exception(response):
                    _move_active_host_to_head(hosts, host)
                    return response
            except requests.exceptions.RequestException:
                pass
        raise errors.ActiveHostNotFound(msg="Could not find active host")


class CdhHdfsConfParser:
    """
    This class serves to automatically extract HDFS cluster information from
    Cloudera configuration files.
    """

    def __init__(self, fs, conf_path=None):
        """
        Args:
            fs (FileSystemClient): The filesystem on which the configuration
                file should be found.
            conf_path (str): The path of the configuration file to be parsed.
        """
        self.fs = fs
        self.conf_path = conf_path or "/etc/hadoop/conf.cloudera.hdfs2/hdfs-site.xml"

    @property
    def config(self):
        if not hasattr(self, "_config"):
            # pylint: disable-next=attribute-defined-outside-init
            self._config = self._get_config()
        return self._config

    def _get_config(self):
        with self.fs.open(self.conf_path) as f:
            d = xml.dom.minidom.parseString(f.read())

        properties = d.getElementsByTagName("property")

        return {
            prop.getElementsByTagName("name")[0]
            .childNodes[0]
            .wholeText: prop.getElementsByTagName("value")[0]
            .childNodes[0]
            .wholeText
            for prop in properties
        }

    @property
    def clusters(self):
        clusters = []
        for key in self.config:
            if key.startswith("dfs.ha.namenodes."):
                clusters.append(key[len("dfs.ha.namenodes.") :])
        return clusters

    def namenodes(self, cluster):
        namenodes = self.config[f"dfs.ha.namenodes.{cluster}"].split(",")
        return [
            self.config[f"dfs.namenode.http-address.{cluster}.{namenode}"]
            for namenode in namenodes
        ]



import re
from collections import OrderedDict


class ParsedNamespaces:
    """
    A namespace parser for DatabaseClient subclasses.

    This is a utility class that encodes the translation of table names in form
    "<name>.<name>...." into instances, given a particular hierarchy of
    namespaces.

    For example, if a particular database has a namespace hierarchy of
    "<database>.<table>", and a specific table name "my_db.my_table"
    was provided, an instance of this class would translate this internally
    into {'database': 'my_db', 'table': 'my_table'}; and expose them by
    <instance>.database == 'my_db' and <instance>.table == 'my_table'.
    Partially complete namespaces (such as 'my_table') will also be parsed,
    interpreting provided names as the least general, and setting the more general
    namespaces to `None` (e.g. in this case, the 'database' namespace to `None`).
    """

    @classmethod
    def from_name(cls, name, namespaces, quote_char='"', separator=".", defaults=None):
        """
        Return an instance of `ParsedNamespaces` from a given name.

        If `name` is an instance of `ParsedNamespaces`, it is checked whether
        the `namespaces` are a subset of the namespaces provided to this
        constructor. If not, a `ValueError` is raised. Note that the quote
        charactors, separators and defaults will of the passed
        `ParsedNamespaces` will be ignored.

        Args:
            name (str, ParsedNamespaces): The name to be parsed.
            namespaces (list<str>): The namespaces into which the name should be
                parsed.
            defaults (None, dict): Default values for namespaces. Note that if a
                default is provided for a namespace, it will only be used if all
                sub-namespaces also resolve to a value (either via defaults or
                by being explicitly passed).
            quote_char (str): The character to used for optional encapsulation
                of namespace names. (default='"')
            separator (str): The character used to separate namespaces.
                (default='.')

        Returns:
            ParsedNamespaces: The `ParsedNamespaces` instance representing the
                parsed value of the nominated name.
        """
        defaults = defaults or {}

        if isinstance(name, ParsedNamespaces):
            extra_namespaces = set(name.namespaces).difference(namespaces)
            if extra_namespaces:
                raise ValueError(
                    f"ParsedNamespace is not encapsulated by the namespaces provided to this constructor. It has extra namespaces: {extra_namespaces}."
                )
            parsed = name.as_dict()

        elif isinstance(name, str):
            namespace_matcher = re.compile(
                r"([^{sep}{qc}]+)|{qc}([^`]*?){qc}".format(
                    qc=re.escape(quote_char), sep=re.escape(separator)
                )
            )

            names = (
                ["".join(t) for t in namespace_matcher.findall(name)] if name else []
            )
            if len(names) > len(namespaces):
                raise ValueError(
                    f"Name '{name}' has too many namespaces. Should be of form: <{'>{separator}<'.join(namespaces)}>."
                )

            parsed = OrderedDict(
                reversed(
                    [
                        (namespace, names.pop() if names else None)
                        for namespace in namespaces[::-1]
                    ]
                )
            )

        else:
            raise ValueError(
                f"Cannot construct `ParsedNamespaces` instance from name of type: `{type(name)}`."
            )

        for namespace in namespaces[::-1]:
            if not parsed.get(namespace) and namespace in defaults:
                parsed[namespace] = defaults[namespace]
            elif not parsed.get(namespace):
                break

        return cls(parsed, quote_char=quote_char, separator=separator)

    def __init__(self, names, namespaces=None, quote_char='"', separator="."):
        if namespaces:
            names = OrderedDict(
                (namespace, names.get(namespace, None)) for namespace in namespaces
            )

        self._names = names
        self._quote_char = quote_char
        self._separator = separator

    def __getattr__(self, name):
        if "_names" in self.__dict__ and name in self._names:
            return self._names[name]
        raise AttributeError(name)

    def __setattr__(self, name, value):
        if "_names" in self.__dict__ and name in self._names:
            self._names[name] = value
        else:
            super().__setattr__(name, value)

    def __bool__(self):
        return bool(self.name)

    def __nonzero__(self):  # Python 2 support for bool
        return bool(self.name)

    @property
    def namespaces(self):
        """list<str> The namespaces parsed in order of most to least specific."""
        return list(self._names)

    @property
    def name(self):
        """str: The full name provided (with quotes)."""
        return self.render()

    @property
    def parent(self):
        """ParsedNamespaces: An instance of `ParsedNamespaces` with the most specific namespace truncated."""
        names = self._names.copy()
        names.popitem()
        return ParsedNamespaces(
            names=names, quote_char=self._quote_char, separator=self._separator
        )

    def as_dict(self):
        """dict: Returns the parsed namespaces as an OrderedDict from most to least general."""
        return self._names

    def render(self, quote_char=None, separator=None):
        if quote_char is None:
            quote_char = self._quote_char
        if separator is None:
            separator = self._separator

        names = [
            self._names[namespace] for namespace, name in self._names.items() if name
        ]
        if len(names) == 0:
            return ""
        return (
            quote_char
            + "{qc}{sep}{qc}".format(qc=quote_char, sep=separator).join(names)
            + quote_char
        )

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Namespace<{self.name}>"



__version__ = '1.0.0'
__author__ = 'John Smith'
__author_email__ = 'john.smith@company.com'

__dependencies__ = [
    # It is recommended to pin omniduct to a specific major version, and manually repin when updates go out
    # Omniduct is quite stable, but this provides a stronger guarantee of stability for users of your package
    'omniduct[ssh,presto,s3]>=1.0.0<1.1.0'
]



from __future__ import print_function

import inspect
import logging
import sys
import time

import progressbar
from decorator import decorate

from .config import config

config.register(
    "logging_level",
    description="Set the default logging level.",
    default=logging.INFO,
    onchange=lambda level: logger.setLevel(level, context="omniduct"),
)


class StatusLogger:
    """
    StatusLogger is a wrapper around `logging.Logger` that allows for consistent
    treatment of logging messages. While not strictly required,
    it simplifies and abstracts the usage of the logging module within omniduct.
    It also adds support for displaying progress bars.

    Instances of StatusLogger are proxies for `logging.Logger` objects, and thus
    inherit the methods and properties of Logger. For example, to log a warning, use:

    >>> StatusLogger().warn('test')

    StatusLogger will automatically detect the context of the logged message, and
    include it in the log messages.
    """

    def __init__(self, auto_scoping=False):
        self.__scopes = []

        ch = LoggingHandler()
        formatter = logging.Formatter(
            "%(levelname)s: %(name)s (%(funcName)s:%(lineno)s): %(message)s"
        )
        ch.setFormatter(formatter)

        self.setLevel(config.logging_level, context="omniduct")
        omniductLogger = self.__get_logger_instance(context="omniduct")
        omniductLogger.addHandler(ch)
        omniductLogger.propagate = False

        self._progress_bar = None

    @property
    def disabled(self):
        return self.__get_logger_instance().disabled

    @disabled.setter
    def disabled(self, disabled):
        self.__get_logger_instance().disabled = disabled

    def _scope_enter(self, name, timed=False, extra=None):
        if config.logging_level < logging.INFO:
            print(
                "\t" * len(self.__scopes) + f"Entering manual scope: {name}",
                file=sys.stderr,
            )
        props = {"name": name}
        if timed:
            props["time"] = time.time()
        if extra is not None:
            props["extra"] = extra
        props["caveats"] = []
        self.__scopes.append(props)

    def _scope_exit(self, success=True):
        if self._progress_bar is not None:
            self.progress(100, complete=True)
        props = self.__scopes[-1]
        if "time" in props:
            logger.warning(
                f"{'Complete' if success else 'Failed'} after {self.__get_time(time.time() - props['time'])} on {time.strftime('%Y-%m-%d')}."
                + (
                    f" CAVEATS: {'; '.join(props['caveats'])}."
                    if props["caveats"]
                    else ""
                )
            )
        scope = self.__scopes.pop()
        if config.logging_level < logging.INFO:
            print(
                "\t" * len(self.__scopes) + f"Exited manual scope: {scope['name']}",
                file=sys.stderr,
            )
        elif "has_logged" in scope:
            if len(self.__scopes) != 0:
                self.current_scope_props["has_logged"] = self.current_scope_props.get(
                    "has_logged"
                ) or props.get("has_logged", False)

    def __get_time(self, seconds):
        m, s = divmod(seconds, 60)
        h, m = divmod(m, 60)

        if h > 0:
            return f"{h:.0f} hrs, {m:.0f} min"
        if m > 0:
            return f"{m:.0f} min, {s:.0f} sec"
        return f"{s:.2f} sec"

    def caveat(self, caveat):
        if len(self.__scopes) == 0:
            self.warning(f"CAVEAT: {caveat}")
        else:
            self.current_scope_props["caveats"].append(caveat)

    @property
    def current_scopes(self):
        """
        The current logger scopes. This is not designed to work with multiple
        threads.
        """
        return detect_scopes()

    @property
    def current_scope_props(self):
        """
        The properties for the most nested manual scope.
        """
        if len(self.__scopes) == 0:
            return None
        return self.__scopes[-1]

    def __get_progress_bar(self, indeterminate=False):
        if self._progress_bar is None:
            if config.logging_level >= logging.INFO:
                prefix = ": ".join(self.current_scopes) + ": "
            else:
                prefix = "\t" * len(self.current_scopes)
            self._progress_bar = progressbar.ProgressBar(
                widgets=[
                    prefix,
                    progressbar.widgets.RotatingMarker()
                    if indeterminate
                    else progressbar.widgets.Bar(),
                    progressbar.widgets.Timer(format=" %(elapsed)s"),
                ],
                redirect_stderr=True,
                redirect_stdout=True,
                max_value=100,
            ).start()

        return self._progress_bar

    def progress(self, progress=None, complete=False, indeterminate=False):
        """
        Set the current progress to `progress`, and if not already showing, display
        a progress bar. If `complete` evaluates to True, then finish displaying the progress.
        """
        complete = complete or (
            self.current_scope_props is None
        )  # Only leave progress bar open if within a scope
        if config.logging_level <= logging.INFO:
            self.__get_progress_bar(indeterminate=indeterminate).update(progress)
            if complete:
                self.__get_progress_bar().finish(end=None)
                self._progress_bar = None

    # Logging emulation

    def __get_logger_instance(self, context=None):
        """
        Get a `logger.Logger` instance for the provided context; inferring the
        context from the runtime stack if the provided context is `None`.
        """
        if context is None:
            try:
                caller = inspect.stack()[2]
                context = inspect.getmodule(caller.frame).__name__
            except:  # pylint: disable=bare-except
                context = "omniduct"
        if context != "omniduct" and not context.startswith("omniduct."):
            context = f"omniduct.external.{context}"
        return logging.getLogger(context)

    def __getattr__(self, name):
        """
        Return the attributes of the wrapped `logging.Logger` instance rather
        than this one (unless the property actually exists in StatusLogger).
        """
        return getattr(self.__get_logger_instance(), name)

    def setLevel(self, level, context=None):
        """
        Add a keyword argument `context` to the standard `setLevel` method, in
        order to allow for fine-grained logging levels, while retaining the
        simplicity of a single "logger" instance.
        """
        self.__get_logger_instance(context).setLevel(level)


def detect_scopes():
    scopes = []
    current_frame = inspect.currentframe()

    while current_frame is not None:
        if current_frame.f_code.co_name == "logging_scope":
            scopes.append(current_frame.f_locals["name"])
        else:
            argvalues = inspect.getargvalues(current_frame)
            if "self" in argvalues.args and getattr(
                argvalues.locals["self"].__class__, "AUTO_LOGGING_SCOPE", False
            ):
                scopes.append(argvalues.locals["self"])
        current_frame = current_frame.f_back

    out_scopes = []
    seen = set()
    for scope in scopes[::-1]:
        if scope not in seen:
            out_scopes.append(
                scope
                if isinstance(scope, str)
                else (
                    getattr(scope, "LOGGING_SCOPE", None)
                    or getattr(scope, "name", None)
                    or scope.__class__.__name__
                )
            )
            seen.add(scope)
    return out_scopes


# pylint: disable-next=abstract-method
class LoggingHandler(logging.Handler):
    """
    An implementation of Logging.Handler to render the logging methods shown in Omniduct and derivatives.
    """

    def __init__(self, level=logging.NOTSET):
        logging.Handler.__init__(self, level=level)
        self.setFormatter(
            logging.Formatter(
                "%(levelname)s: %(name)s (%(funcName)s:%(lineno)s): %(message)s"
            )
        )

    def format_simple(self, record):
        return f"{record.getMessage()}"

    def handle(self, record):
        try:
            scopes = logger.current_scopes
        except:  # pylint: disable=bare-except
            scopes = []

        if config.logging_level < logging.INFO:  # Print everything verbosely
            prefix = "\t" * len(scopes)
            self._overwrite(
                prefix + self.format(record), overwritable=False, truncate=False
            )
        else:
            prefix = ""
            important = (
                record.levelno >= logging.WARNING
                or logger._progress_bar is not None
                or len(scopes) == 0
            )

            if len(scopes) > 0:
                prefix = ": ".join(scopes) + ": "
                if logger.current_scope_props is not None:
                    logger.current_scope_props["has_logged"] = True

            self._overwrite(
                prefix + self.format_simple(record),
                overwritable=not important,
                truncate=not important,
            )

        sys.stderr.flush()

    def _overwrite(self, text, overwritable=True, truncate=True, file=sys.stderr):
        w, _ = progressbar.utils.get_terminal_size()
        file.write("\r" + " " * w + "\r")  # Clear current line
        if overwritable:
            text.replace("\n", " ")
        if truncate:
            if len(text) > w:
                text = text[: w - 3] + "..."
        if not overwritable:
            text += "\n"
        file.write(text)


def logging_scope(name, *wargs, **wkwargs):
    """
    A decorator to add the decorated function as a new logging scope, with name `name`.
    All additional arguments are passed to `StatusLogger._scope_enter`. Current
    supported keyword arguments are "timed", in which case when the scope closes,
    the duration of the call is shown.
    """

    def logging_scope(func, *args, **kwargs):
        logger._scope_enter(name, *wargs, **wkwargs)
        success = True
        try:
            f = func(*args, **kwargs)
            return f
        except Exception:  # pylint: disable=broad-exception-caught
            success = False
            raise
        finally:
            logger._scope_exit(success)

    return lambda func: decorate(func, logging_scope)


logger = StatusLogger()



import os
import sys

try:
    from ._version_info import __version__, __version_tuple__
except ImportError:
    __version__ = "unknown"
    __version_tuple__ = (0, 0, 0, "+unknown")

__all__ = [
    "__author__",
    "__author_email__",
    "__version__",
    "__version_tuple__",
    "__logo__",
    "__docs_url__",
]

__author__ = "Matthew Wardrop, Dan Frank"
__author_email__ = "mpwardrop@gmail.com, danfrankj@gmail.com"
__logo__ = (
    os.path.join(os.path.dirname(__file__), "logo.png")
    if "__file__" in globals()
    else None
)
__docs_url__ = "https://omniduct.readthedocs.io/"


# These are the core dependencies, and should not include those which are used only in handling specific protocols.
# Order matters since installation happens from the end of the list
__dependencies__ = [
    "interface_meta>=1.1.0,<2",  # Metaclass for creating an extensible well-documented architecture
    "pyyaml",  # YAML configuration parsing
    "decorator",  # Decorators used by caching and documentation routines
    "progressbar2>=3.30.0",  # Support for progressbars in logging routines
    "wrapt",  # Object proxying for conveniently exposing ducts in registry
    # Database querying libraries
    "jinja2",  # Templating support in databases
    "pandas>=0.20.3",  # Various results including database queries are returned as pandas dataframes
    "sqlparse",  # Neatening of SQL based queries (mainly to avoid missing the cache)
    "sqlalchemy",  # Various integration endpoints in the database stack
    # Utility libraries
    "python-dateutil",  # Used for its `relativedelta` class for Cache instances
    "lazy-object-proxy",  # Schema traversal
]

__optional_dependencies__ = {
    # Databases
    "druid": [
        "pydruid>=0.4.0",  # Primary client
    ],
    "hiveserver2": [
        "pyhive[hive]>=0.4",  # Primary client
        "thrift>=0.10.0",  # Thrift dependency which seems not to be installed with upstream deps
    ],
    "presto": [
        "pyhive[presto]>=0.4",  # Primary client
    ],
    "pyspark": [
        "pyspark",  # Primary client
    ],
    "snowflake": [
        "snowflake-sqlalchemy",
    ],
    "exasol": ["pyexasol"] if sys.version_info.major > 2 else [],
    # Filesystems
    "webhdfs": [
        "pywebhdfs",  # Primary client
        "requests",  # For rerouting redirect queries to our port-forwarded services
    ],
    "s3": [
        "boto3",  # AWS client library
    ],
    # Remotes
    "ssh": [
        "pexpect",  # Command line handling (including smartcard activation)
    ],
    "ssh_paramiko": [
        "paramiko",  # Primary client
        "pexpect",  # Command line handling (including smartcard activation)
    ],
    # Rest clients
    "rest": [
        "requests",  # Library to handle underlying REST queries
    ],
    # Documentation requirements
    "docs": [
        "sphinx",  # The documentation engine
        "sphinx_autobuild",  # A Sphinx plugin used during development of docs
        "sphinx_rtd_theme",  # The Spinx theme used by the docs
    ],
    "test": [
        "nose",  # test runner
        "mock",  # mocking
        "pyfakefs",  # mock filesystem
        "coverage",  # test coverage monitoring
        "flake8",  # Code linting
    ],
}
__optional_dependencies__["all"] = [
    dep for deps in __optional_dependencies__.values() for dep in deps
]



import importlib
import re
from typing import Optional

import packaging.requirements

from omniduct._version import __optional_dependencies__
from omniduct.utils.debug import logger


def get_package_version(package_name: str) -> Optional[str]:
    """
    Return the version of the given package, or None if the package is not
    installed.
    """
    try:  # Python 3.8+
        import importlib.metadata

        return importlib.metadata.version(package_name)
    except ImportError:  # Python <3.12
        import pkg_resources

        return pkg_resources.get_distribution(package_name).version


def check_dependencies(protocols, message=None):
    if protocols is None:
        return
    dependencies = []
    for protocol in protocols:
        dependencies.extend(__optional_dependencies__.get(protocol, []))
    missing_deps = []
    warning_deps = {}

    for dep in dependencies:
        m = re.match("^[a-z_][a-z0-9]*", dep)
        if not m:
            logger.warning(f"Invalid dependency requested: {dep}")

        package_name = m.group(0)
        accept_any_version = package_name == dep

        dep_req = packaging.requirements.Requirement(dep)
        package_name = dep_req.name
        package_version = get_package_version(package_name)

        if package_version is None:
            # Some packages may be available, but not installed. If so, we
            # should accept them with warnings (if version specified in dep).
            try:
                importlib.import_module(package_name)
                if not accept_any_version:
                    warning_deps[dep] = f"{package_name}==<not installed>"
            except ModuleNotFoundError:
                missing_deps.append(dep)
        elif dep_req.specifier and not dep_req.specifier.contains(package_version):
            warning_deps[dep] = f"{package_name}=={package_version}"

    if warning_deps:
        message = "You may have some outdated packages:\n"
        for key in sorted(warning_deps):
            message += f"\t- Want {key}, found {warning_deps[key]}"
        logger.warning(message)
    if missing_deps:
        message = (
            message or "Whoops! You do not seem to have all the dependencies required."
        )
        fix = (
            "You can fix this by running:\n\n"
            f"\tpip install --upgrade {' '.join(missing_deps)}\n\n"
            "Note: Depending on your system's installation of Python, you may "
            "need to use `pip3` instead of `pip`."
        )
        raise RuntimeError("\n\n".join([message, fix]))
